import _ from 'lodash';

import * as audit from './clients/audit';
import * as generic from './clients/generic';
import * as list from './clients/list';
import * as search from './clients/search';
import * as members from './clients/members';
import * as layerMembers from './clients/layer_members';
import * as graph from './clients/graph';

import {RManager} from './clients/request';

import constants from './constants';
import {readCookie} from './utils';

let rmanager;

function getFullId(kind, id) {
    return window.encodeURIComponent(
        `${rmanager.account}:${kind}:${id}`
    );
}

export default {
    getRManager() {
        if (_.isUndefined(rmanager)) {
            rmanager = new RManager(this.flux, readCookie('conjur_configuration'));
        }

        return rmanager;
    },

    getFullId(kind, id) {
        return getFullId(kind, id);
    },

    flashMessage(msg, level) {
        this.dispatch(constants.FLASH_MESSAGE, {msg: msg, level: level});
    },

    flashMessageClear() {
        this.dispatch(constants.FLASH_MESSAGE_CLEAR);
    },

    routes: {
        transition(params) {
            this.dispatch(constants.ROUTES.TRANSITION, {params: params});
        }
    },

    search(query) {
        this.dispatch(constants.SEARCH.LOAD, {id: query});

        search
            .fetch(rmanager, query)
            .then(
                res => {
                    const payload = {id: query, type: 'SEARCH_RESULT', res: res};

                    this.dispatch(
                        constants.SEARCH.LOAD_SUCCESS,
                        payload
                    );
                },
                () => {
                    this.dispatch(constants.SEARCH.LOAD_FAIL, {id: query, type: 'SEARCH_RESULT'});
                });
    },

    audit: {
        loadAll(startFlow) {
            this.dispatch(constants.AUDIT.LOAD);

            audit
                .fetch(rmanager, audit.c.AUDIT_ALL, null)
                .then(
                    res => {
                        const payload = {res: res, startFlow: startFlow};

                        this.dispatch(
                            constants.AUDIT.LOAD_SUCCESS,
                            payload
                        );
                    },
                    () => {
                        this.dispatch(constants.AUDIT.LOAD_FAIL);
                    });
        },

        loadFlow(/*lastEventId*/) {
            // let eventSource = audit.fetch(audit.c.AUDIT_FLOW, null);

            // eventSource.onmessage = event => {
            //     const data = JSON.parse(event.data);

            //     if (data.id > lastEventId) {
            //         const payload = {event: data};

            //         this.dispatch(
            //             constants.AUDIT.LOAD_SUCCESS,
            //             payload);
            //     }
            // };

            // eventSource.onerror = () => {
            //     // this.dispatch(constants.AUDIT.LOAD_FAIL);
            // };
        },

        loadForRole(kind, id) {
            this.dispatch(constants.AUDIT.LOAD);

            audit
                .fetch(
                    rmanager,
                    audit.c.AUDIT_ROLE,
                    getFullId(kind, id)
                )
                .then(
                    res => {
                        const payload = {res: res};

                        this.dispatch(
                            constants.AUDIT.LOAD_SUCCESS,
                            payload
                        );
                    },
                    () => {
                        this.dispatch(constants.AUDIT.LOAD_FAIL, {kind: kind, id: id});
                    });
        },

        loadForResource(kind, id) {
            this.dispatch(constants.AUDIT.LOAD);

            audit
                .fetch(
                    rmanager,
                    audit.c.AUDIT_RESOURCE,
                    getFullId(kind, id)
                )
                .then(
                    res => {
                        const payload = {res: res};

                        this.dispatch(
                            constants.AUDIT.LOAD_SUCCESS,
                            payload
                        );
                    },
                    () => {
                        this.dispatch(constants.AUDIT.LOAD_FAIL, {kind: kind, id: id});
                    });
        }
    },

    resources: {
        load(kind) {
            this.dispatch(constants.RESOURCES.LOAD);

            list
                .fetch(rmanager, kind)
                .then(
                    res => {
                        this.dispatch(constants.RESOURCES.LOAD_SUCCESS, {res: res});
                    },
                    () => {
                        this.dispatch(constants.RESOURCES.LOAD_FAIL, {id: kind});
                    });
        },

        loadOne(kind, id) {
            this.dispatch(constants.RESOURCE.LOAD);

            list
                .fetchOne(rmanager, kind, id)
                .then(
                    res => {
                        this.dispatch(constants.RESOURCE.LOAD_SUCCESS, {kind: kind, id: id, res: res});
                    },
                    () => {
                        this.dispatch(constants.RESOURCE.LOAD_FAIL, {kind: kind, id: id});
                    });
        }
    },

    graph: {
        load(kind, id) {
            this.dispatch(constants.GRAPH.LOAD);

            graph
                .fetch(
                    rmanager,
                    getFullId(kind, id)
                )
                .then(
                    res => {
                        const payload = {kind: kind, id: id, res: res};

                        this.dispatch(
                            constants.GRAPH.LOAD_SUCCESS,
                            payload
                        );
                    },
                    () => {
                        this.dispatch(constants.GRAPH.LOAD_FAIL, {id: id});
                    });
        }
    },

    user: {
        load(id) {
            this.dispatch(constants.USER.LOAD, {id: id});

            [
                generic.c.ATTRIBUTES,
                generic.c.OWNED_RESOURCES,
                generic.c.ALL_ROLES,
                generic.c.PUBLIC_KEYS
            ].forEach((type) => {
                this.dispatch(constants.USER.LOAD, {id: id, type: type});

                generic
                    .fetch(rmanager, type, 'user', id)
                    .then(
                        res => {
                            const payload = {id: id, type: type, res: res};

                            this.dispatch(
                                constants.USER.LOAD_SUCCESS,
                                payload
                            );
                        },
                        () => {
                            this.dispatch(constants.USER.LOAD_FAIL, {type: type});
                        });
            });

            this.dispatch(constants.USER.LOAD, {id: id, type: 'RESOURCES'});

            list
                .fetchResources(rmanager, getFullId('user', id))
                .then(
                    res => {
                        const payload = {id: id, type: 'RESOURCES', res: res};

                        this.dispatch(
                            constants.USER.LOAD_SUCCESS,
                            payload
                        );
                    },
                    () => {
                        this.dispatch(constants.USER.LOAD_FAIL, {type: 'RESOURCES'});
                    });
        }
    },

    group: {
        load(id) {
            this.dispatch(constants.GROUP.LOAD, {id: id});

            [
                generic.c.ATTRIBUTES,
                generic.c.OWNED_RESOURCES,
                generic.c.ALL_ROLES
            ].forEach((type) => {
                this.dispatch(constants.GROUP.LOAD, {id: id, type: type});

                generic
                    .fetch(rmanager, type, 'group', id)
                    .then(
                        res => {
                            const payload = {id: id, type: type, res: res};

                            this.dispatch(
                                constants.GROUP.LOAD_SUCCESS,
                                payload
                            );
                        },
                        () => {
                            this.dispatch(constants.GROUP.LOAD_FAIL, {type: type});
                        });
            });

            this.dispatch(constants.GROUP.LOAD, {id: id, type: 'GROUP_ROLES'});

            list
                .fetchGroupRoles(rmanager, id)
                .then(
                    res => {
                        const payload = {id: id, type: 'GROUP_ROLES', res: res};

                        this.dispatch(
                            constants.GROUP.LOAD_SUCCESS,
                            payload
                        );
                    },
                    () => {
                        this.dispatch(constants.GROUP.LOAD_FAIL, {type: 'GROUP_ROLES'});
                    });

            this.dispatch(constants.GROUP.LOAD, {id: id, type: 'RESOURCES'});

            list
                .fetchResources(rmanager, getFullId('group', id))
                .then(
                    res => {
                        const payload = {id: id, type: 'RESOURCES', res: res};

                        this.dispatch(
                            constants.GROUP.LOAD_SUCCESS,
                            payload
                        );
                    },
                    () => {
                        this.dispatch(constants.GROUP.LOAD_FAIL, {type: 'RESOURCES'});
                    });
        }
    },

    host: {
        load(id) {
            this.dispatch(constants.HOST.LOAD, {id: id});

            [
                generic.c.ATTRIBUTES,
                generic.c.OWNED_RESOURCES,
                generic.c.ALL_ROLES
            ].forEach((type) => {
                this.dispatch(constants.HOST.LOAD, {id: id, type: type});

                generic
                    .fetch(rmanager, type, 'host', id)
                    .then(
                        res => {
                            const payload = {id: id, type: type, res: res};

                            this.dispatch(
                                constants.HOST.LOAD_SUCCESS,
                                payload
                            );
                        },
                        () => {
                            this.dispatch(constants.HOST.LOAD_FAIL, {type: type});
                        });
            });

            [
                members.c.MEMBERS_EXECUTE,
                members.c.MEMBERS_UPDATE
            ].forEach((type) => {
                this.dispatch(constants.HOST.LOAD, {id: id, type: type});

                members
                    .fetch(rmanager, constants.HOST.ASSET_TYPE, type, id)
                    .then(
                        res => {
                            const payload = {id: id, type: type, res: res};

                            this.dispatch(
                                constants.HOST.LOAD_SUCCESS,
                                payload
                            );
                        },
                        () => {
                            this.dispatch(constants.HOST.LOAD_FAIL, {type: type});
                        });
            });

            this.dispatch(constants.HOST.LOAD, {id: id, type: 'RESOURCES'});

            list
                .fetchResources(rmanager, getFullId('host', id))
                .then(
                    res => {
                        const payload = {id: id, type: 'RESOURCES', res: res};

                        this.dispatch(
                            constants.HOST.LOAD_SUCCESS,
                            payload
                        );
                    },
                    () => {
                        this.dispatch(constants.HOST.LOAD_FAIL, {type: 'RESOURCES'});
                    });
        }
    },

    layer: {
        load(id) {
            this.dispatch(constants.LAYER.LOAD, {id: id});

            [
                generic.c.ATTRIBUTES,
                generic.c.OWNED_RESOURCES,
                generic.c.ALL_ROLES
            ].forEach((type) => {
                this.dispatch(constants.LAYER.LOAD, {id: id, type: type});

                generic
                    .fetch(rmanager, type, 'layer', id)
                    .then(
                        res => {
                            const payload = {id: id, type: type, res: res};

                            this.dispatch(
                                constants.LAYER.LOAD_SUCCESS,
                                payload
                            );
                        },
                        () => {
                            this.dispatch(constants.LAYER.LOAD_FAIL, {type: type});
                        });
            });

            [
                layerMembers.c.LAYER_MEMBERS_USE,
                layerMembers.c.LAYER_MEMBERS_ADMIN
            ].forEach((type) => {
                this.dispatch(constants.LAYER.LOAD, {id: id, type: type});

                layerMembers
                    .fetch(rmanager, type, id)
                    .then(
                        res => {
                            const payload = {id: id, type: type, res: res};

                            this.dispatch(
                                constants.LAYER.LOAD_SUCCESS,
                                payload
                            );
                        },
                        () => {
                            this.dispatch(constants.LAYER.LOAD_FAIL, {type: type});
                        });
            });

            this.dispatch(constants.LAYER.LOAD, {id: id, type: 'RESOURCES'});

            list
                .fetchResources(rmanager, getFullId('layer', id))
                .then(
                    res => {
                        const payload = {id: id, type: 'RESOURCES', res: res};

                        this.dispatch(
                            constants.LAYER.LOAD_SUCCESS,
                            payload
                        );
                    },
                    () => {
                        this.dispatch(constants.LAYER.LOAD_FAIL, {type: 'RESOURCES'});
                    });
        }
    },

    variable: {
        load(id) {
            this.dispatch(constants.VARIABLE.LOAD, {id: id});

            [
                generic.c.ATTRIBUTES
            ].forEach((type) => {
                this.dispatch(constants.VARIABLE.LOAD, {id: id, type: type});

                generic
                    .fetch(rmanager, type, 'variable', id)
                    .then(
                        res => {
                            const payload = {id: id, type: type, res: res};

                            this.dispatch(
                                constants.VARIABLE.LOAD_SUCCESS,
                                payload
                            );
                        },
                        () => {
                            this.dispatch(constants.VARIABLE.LOAD_FAIL, {type: type});
                        });
            });

            [
                members.c.MEMBERS_EXECUTE,
                members.c.MEMBERS_UPDATE
            ].forEach((type) => {
                this.dispatch(constants.VARIABLE.LOAD, {id: id, type: type});

                members
                    .fetch(rmanager, constants.VARIABLE.ASSET_TYPE, type, id)
                    .then(
                        res => {
                            const payload = {id: id, type: type, res: res};

                            this.dispatch(
                                constants.VARIABLE.LOAD_SUCCESS,
                                payload
                            );
                        },
                        () => {
                            this.dispatch(constants.VARIABLE.LOAD_FAIL, {type: type});
                        });
            });
        }
    },

    policy: {
        load(id) {
            this.dispatch(constants.POLICY.LOAD, {id: id});

            [
                generic.c.OWNED_RESOURCES,
                generic.c.ALL_ROLES
            ].forEach((type) => {
                this.dispatch(constants.POLICY.LOAD, {id: id, type: type});

                generic
                    .fetch(rmanager, type, 'policy', id)
                    .then(
                        res => {
                            const payload = {id: id, type: type, res: res};

                            this.dispatch(
                                constants.POLICY.LOAD_SUCCESS,
                                payload
                            );
                        },
                        () => {
                            this.dispatch(constants.POLICY.LOAD_FAIL, {type: type});
                        });
            });

            this.dispatch(constants.POLICY.LOAD, {id: id, type: 'RESOURCES'});

            list
                .fetchResources(rmanager, getFullId('policy', id))
                .then(
                    res => {
                        const payload = {id: id, type: 'RESOURCES', res: res};

                        this.dispatch(
                            constants.POLICY.LOAD_SUCCESS,
                            payload
                        );
                    },
                    () => {
                        this.dispatch(constants.POLICY.LOAD_FAIL, {type: 'RESOURCES'});
                    });
        }
    }
};
