import {Buffer} from 'buffer';
import _ from 'lodash';
import superagent from 'superagent';
import moment from 'moment';

function PRequest(method, url, rmanager) {
    superagent.Request.apply(this, [method, url]);

    this.rmanager = rmanager;
    this._highPriority = false;
}

PRequest.prototype = Object.create(superagent.Request.prototype);

PRequest.prototype.highPriority = function() {
    this._highPriority = true;
    return this;
};

PRequest.prototype.isHighPriority = function() {
    return this._highPriority;
};

PRequest.prototype.end = function() {
    let _super = superagent.Request.prototype.end,
        self = this;

    return new Promise(function(accept, reject) {
        self.rmanager.placeOnWaitQ(_super, self, accept, reject);
    });
};

var configProps = {
    _conf: {
        value: {
            applianceUrl: null,
            account: null,
            username: null,
            apiKey: null,
            token: null,
            tokenExpire: null,
            requestToken: null
        }
    }
};

/**
 *
 * RManagerNoopConfig provides a no-op configuration for
 * RManager. Allows in-memory storage of attributes, always shows the
 * user as logged in, disallows logout. Used when the UI is launched
 * in plugin mode.
 *
 */
function RManagerNoopConfig(configCookie) {
    var cookie = JSON.parse(configCookie);
    this.account = cookie.account;
    this.applianceUrl = `http://${document.location.host}`;
    this.username = cookie.login;
}

Object.defineProperties(RManagerNoopConfig.prototype, configProps);
Object.defineProperties(RManagerNoopConfig.prototype, {
    applianceUrl: {
        get: function getApplianceUrl() {
            return this._conf.applianceUrl;
        },
        set: function setApplianceUrl(applianceUrl) {
            this._conf.applianceUrl = applianceUrl;
        }
    },
    account: {
        get: function getAccount() {
            return this._conf.account;
        },
        set: function setAccount(account) {
            this._conf.account = account;
        }
    },
    username: {
        get: function getUsername() {
            return this._conf.username;
        },
        set: function setUsername(username) {
            this._conf.username = username;
        }
    },
    apiKey: {
        get: function getApiKey() {
            return this._conf.apiKey;
        },
        set: function setApiKey(apiKey) {
            this._conf.apiKey = apiKey;
        }
    },
    token: {
        get: function getToken() {
            return this._conf.token;
        },
        set: function setToken(token) {
            this._conf.token = token;
        }
    },
    tokenExpire: {
        get: function getTokenExpire() {
            return this._conf.tokenExpire;
        },
        set: function setTokenExpire(tokenExpire) {
            this._conf.tokenExpire = tokenExpire;
        }
    },
    isLogged: {
        get: function() {
            return true;
        }
    },
    allowLogout: {
        get: function() {
            return false;
        }
    }
});

RManagerNoopConfig.prototype._saveState = function() {};
RManagerNoopConfig.prototype._restoreState = function() {};
RManagerNoopConfig.prototype._removeStoredState = function() {};

/**
 *
 * RManagerLocalStorageConfig provides a local-storage-backed
 * configuration for RManager. Persists the provided attributes to
 * local storage, manages refreshing the user's token, shows the user
 * as logged in while he has an active token, allows logout.
 *
 * Used when connecting to server-side UI.
 */

function RManagerLocalStorageConfig() {
    this._timer = null;
}

Object.defineProperties(RManagerLocalStorageConfig.prototype, configProps);
Object.defineProperties(RManagerLocalStorageConfig.prototype, {
    applianceUrl: {
        get: function getApplianceUrl() {
            return this._conf.applianceUrl;
        },
        set: function setApplianceUrl(applianceUrl) {
            this._conf.applianceUrl = applianceUrl;
            this.account = null;
        }
    },
    account: {
        get: function getAccount() {
            return this._conf.account;
        },
        set: function setAccount(account) {
            this._conf.account = account;
            this.username = null;
        }
    },
    username: {
        get: function getUsername() {
            return this._conf.username;
        },
        set: function setUsername(username) {
            this._conf.username = username;
            this.apiKey = null;
        }
    },
    apiKey: {
        get: function getApiKey() {
            return this._conf.apiKey;
        },
        set: function setApiKey(apiKey) {
            this._conf.apiKey = apiKey;
            this.token = null;
        }
    },
    token: {
        get: function getToken() {
            return this._conf.token;
        },
        set: function setToken(token) {
            if (_.isNull(token)) {
                this._conf.token = null;
            } else if (_.isString(token)) {
                this._conf.token = token;
            } else {
                this._conf.token = new Buffer(JSON.stringify(token)).toString('base64');
            }

            this.tokenExpire = null;
        }
    },
    tokenExpire: {
        get: function getTokenExpire() {
            return this._conf.tokenExpire;
        },
        set: function setTokenExpire(tokenExpire) {
            this._stopTokenRefreshTimeout();

            if (_.isNull(tokenExpire)) {
                this._conf.tokenExpire = null;
            } else {
                const expirationDate = tokenExpire + 300000,
                      timeout = expirationDate - moment.utc();

                this._startTokenRefreshTimeout(timeout > 0 ? timeout : 1);
                this._conf.tokenExpire = expirationDate;
            }
        }
    },
    isLogged: {
        get: function isLogged() {
            let diff = this.tokenExpire - moment.utc();

            if (_.isString(this.applianceUrl) && _.isString(this.account) &&
                _.isString(this.username) && _.isString(this.apiKey) &&
                _.isString(this.token) && _.isNumber(this.tokenExpire) &&
                (diff > 0)) {
                return true;
            } else {
                return false;
            }
        }
    },
    allowLogout: {
        get: function() {
            return true;
        }
    }
});

RManagerLocalStorageConfig.prototype._saveState = function() {
    if (this.isLogged) {
        let state = {
            applianceUrl: this.applianceUrl,
            account: this.account,
            username: this.username,
            apiKey: this.apiKey,
            token: this.token,
            tokenExpire: this.tokenExpire
        };

        if (window.localStorage) {
            window.localStorage.conjur = JSON.stringify(state);
        }
    }
};

RManagerLocalStorageConfig.prototype._restoreState = function() {
    let state;

    try {
        if (window.localStorage) {
            state = JSON.parse(window.localStorage.conjur);
        }
    } catch(e) {
        this._removeStoredState();
    }

    if (_.isObject(state)) {
        let diff = state.tokenExpire - moment.utc();

        if (diff > 1000) { // expired
            this.applianceUrl = state.applianceUrl;
            this.account = state.account;
            this.username = state.username;
            this.apiKey = state.apiKey;
            this.token = state.token;
            this.tokenExpire = state.tokenExpire - 300000;
        } else {
            this._removeStoredState();
        }
    }
};

RManagerLocalStorageConfig.prototype._removeStoredState = function() {
    if (window.localStorage) {
        window.localStorage.conjur = null;
    }
};

RManagerLocalStorageConfig.prototype._startTokenRefreshTimeout = function(timeout) {
    this._timer = window.setTimeout(() => {
        this.requestToken();
    }, timeout);
};

RManagerLocalStorageConfig.prototype._stopTokenRefreshTimeout = function() {
    window.clearTimeout(this._timer);
};

export function RManager(flux, configCookie) {
    this.flux = flux;

    this._config = configCookie? new RManagerNoopConfig(configCookie) : new RManagerLocalStorageConfig();
    this._config._restoreState();
    this._config.requestToken = _.bind(this.requestToken, this);

    this._waitQueue = [];
    this._runPool = {};
    this._nextRequestId = 1;
}

Object.defineProperties(RManager.prototype, {
    account: {
        get: function() {
            return this._config.account;
        },
        set: function(account) {
            this._config.account = account;
        }
    },
    username: {
        get: function() {
            return this._config.username;
        },
        set: function(u) {
            this._config.username = u;
        }
    },
    isLogged: {
        get: function() {
            return this._config.isLogged;
        }
    },
    allowLogout: {
        get: function() {
            return this._config.allowLogout;
        }
    }
});

RManager.prototype.placeOnWaitQ = function(run, request, accept, reject) {
    let self = this;

    this._waitQueue.push({
        request: request,
        run: function(index) {
            run.call(request, function(err, value) {
                if (err || value.error) {
                    if (value && value.error) {
                        self.flux.actions.flashMessage(value.error.message);
                    } else {
                        self.flux.actions.flashMessage(err.message);
                    }

                    reject(value);
                } else {
                    accept(value);
                }

                self.removeFromRunPool(index);
            });
        }
    });

    this._notify();
};

RManager.prototype.emptyWaitQ = function() {
    this._waitQueue = [];
};

RManager.prototype.takeFromWaitQ = function() {
    return this._waitQueue.shift();
};

RManager.prototype.pushBackToWaitQ = function(request) {
    this._waitQueue.unshift(request);
};

RManager.prototype.isWaitQEmpty = function() {
    return this._waitQueue.length === 0;
};

RManager.prototype.isFirstOnWaitQHighPriority = function() {
    return this._waitQueue[0] && this._waitQueue[0].request.isHighPriority();
};

RManager.prototype.placeInRunPool = function(request) {
    const requestId = this._nextRequestId;
    this._nextRequestId += 1;

    this._runPool[requestId] = request;

    return requestId;
};

RManager.prototype.removeFromRunPool = function(requestId) {
    delete this._runPool[requestId];
};

RManager.prototype.isHighPriorityInRunPool = function() {
    return _.isObject(_.find(this._runPool, (e) => {
        return _.isObject(e) && e.request.isHighPriority();
    }));
};

RManager.prototype._notify = function() {
    // inspect run q for request to refresh token
    if (this.isHighPriorityInRunPool()) {
        // if yes - end method
        return;
        // if no - go forward
        // if waitq is not empty - go forward
    } else if (!this.isWaitQEmpty()) {
        // if first on waitq is refresh token request
        // run and place it on runq - end
        if (this.isFirstOnWaitQHighPriority()) {
            let request = this.takeFromWaitQ(),
                requestId = this.placeInRunPool(request);

            request.run(requestId);
        } else {
            // take from waitq util request to refresh token endpoint
            // run them all and place on runq - end
            let request,
                requestId;

            for (;;) {
                request = this.takeFromWaitQ();

                if (_.isUndefined(request) || _.isNull(request)) {
                    break;
                }

                if (request.request.isHighPriority()) {
                    this.pushBackToWaitQ(request);

                    break;
                }

                requestId = this.placeInRunPool(request);

                request.run(requestId);
            }
        }
    }
};

RManager.prototype.getRequest = function(method, path) {
    let url = path;

    if (_.isString(this._config.applianceUrl)) {
        url = `${this._config.applianceUrl}${url}`;
    }

    return new PRequest(method, url, this);
};

RManager.prototype.callServer = function(verb, path) {
    let req = this.getRequest(
        verb,
        path.replace('\${account}', window.encodeURIComponent(this._config.account))
    );

    if (typeof this._config.token === 'string') {
        req.set('Authorization', `Token token="${this._config.token}"`);
    }

    return req;
};

RManager.prototype.login = function(applianceUrl, username, password) {
    return this.requestAccount(applianceUrl)
        .then(() => {
            return this.requestApiKey(username, password)
                .then(() => {
                    return this._config.requestToken();
                });
        })
        .then(
            () => {
                return true;
            },
            () => {
                return false;
            });
};

RManager.prototype.logout = function() {
    this._config.applianceUrl = null;
    this._config._removeStoredState();
    this.emptyWaitQ();
};

RManager.prototype.requestAccount = function(applianceUrl) {
    return this
        .getRequest('GET', '/api/info')
        .timeout(30000)
        .highPriority()
        .end()
        .then(res => {
            this._config.applianceUrl = applianceUrl;
            this._config.account = res.body.account;
        });
};

RManager.prototype.requestApiKey = function(username, password) {
    return this
        .getRequest('GET', '/api/authn/users/login')
        .auth(username, password)
        .timeout(30000)
        .highPriority()
        .set('Accept', 'text/plain')
        .end()
        .then(res => {
            this._config.username = username;
            this._config.apiKey = res.text;
        });
};

RManager.prototype.requestToken = function() {
    return this
        .getRequest('POST', `/api/authn/users/${this._config.username}/authenticate`)
        .send(this._config.apiKey)
        .timeout(30000)
        .highPriority()
        .end()
        .then(res => {
            this._config.token = res.body;
            this._config.tokenExpire = moment.utc(res.body.timestamp, 'YYYY-MM-DD hh:mm:ss UTC');

            if (this._config.isLogged) {
                this._config._saveState();
            }
        });
};
