import Fluxxor from 'fluxxor';
import SortedSet from 'collections/sorted-set';
import _ from 'lodash';

import constants from '../components/audit/constants';
import c from '../constants';

/**
 * AuditStore stores audit events
 *
 * Use SortedSet as internal data structure.
 */
module.exports = Fluxxor.createStore({
    initialize() {
        this.loading = 0;
        this.events = new SortedSet([], function(a, b) {
            return b.id === a.id;
        }, function(a, b) {
            if (a.id < b.id) {
                return -1;
            }

            if (a.id > b.id) {
                return 1;
            }

            return 0;
        });

        this.bindActions(
            c.AUDIT.LOAD, this._onLoad,
            c.AUDIT.LOAD_SUCCESS, this._onLoadSuccess,
            c.AUDIT.LOAD_FAIL, this._onLoadFail
        );
    },

    _incLoading() {
        this.loading = this.loading + 1;
    },

    _decLoading() {
        this.loading = this.loading - 1 < 0 ? 0 : this.loading - 1;
    },

    _isLoading() {
        return Boolean(this.loading);
    },

    _filterEvent(e) {
        // remove internal resources creation
        if (_.isString(e.role) && (e.role.split(':')[1] === '@')) {
            return false;
        }

        if (_.isString(e.resource) && (e.resource.split(':')[1] === '@')) {
            return false;
        }

        if (_.isString(e.resource) && (e.resource.split(':')[1] === 'secret')) {
            return false;
        }

        if (_.isString(e.grantee) && (e.grantee.split(':')[1] === '@')) {
            return false;
        }

        // hide automated creation of roles corresponding to resources of known type
        if ((e.kind === 'role') && (e.action === 'create')) {
            var roleKind = e.role.split(':')[1];

            if (_.includes(constants.knownRolsourceTypes, roleKind)) {
                return false;
            }
        }

        // hide new event `role:index` / `all_roles`, which is used by role graph only
        if ((e.kind === 'role') && (e.action === 'index')) {
            return false;
        }

        if ((e.kind === 'role') && (e.action === 'all_roles')) {
            return false;
        }

        return true;
    },

    _onLoad() {
        this._incLoading();
        this.emit('change');
    },

    _onLoadSuccess(payload) {
        if (_.isObject(payload.event)) {
            this.events.addEach(_.filter([payload.event], this._filterEvent));
            this.emit('change');
        } else if (_.isArray(payload.res.body)) {
            this._decLoading();
            this.events.addEach(_.filter(payload.res.body, this._filterEvent));
            this.emit('change');
        }

        if (payload.startFlow === true) {
            this.flux.actions.audit.loadFlow(this.events.findGreatest().value.id);
        }
    },

    _onLoadFail() {
        this._decLoading();
        this.emit('change');
    },

    /**
     * Return all exist events in store
     */
    getAllEvents() {
        return {
            loading: Boolean(this.loading),
            data: this.events.toArray()
        };
    },

    getEventFor(kind, id) {
        var fullId = window.decodeURIComponent(this.flux.actions.getFullId(kind, id));

        return {
            loading: this._isLoading(),
            data: this.events.filter(function(event) {

                var resourceId = event.resource;
                if ((!event.resource) && (event.resources)) {
                    resourceId = event.resources[0];
                }
                var roleId = event.role;
                if ((!event.role) && (event.roles)) {
                    roleId = event.roles[0];
                }

                return roleId === fullId || resourceId === fullId;
            }).toArray()
        };
    }
});
